/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

File: csmrc.pl
Author: Temur Kutsia
Date: 28.11.2005
Implementation of the Context Sequence Matching Algorithm with regular
constraints.
Copyright (c) 2005 Temur Kutsia

This program will run on the following Prolog dialects:
SWI-Prolog, GNU-Prolog, BinProlog, LPA Win-Prolog. It does not run
on YAP Prolog.

=========================================================================

Input syntax:
  Context and sequence variables:
      Identifiers that start with lower case letters and end with double
      underscores. Anonymous context and sequence variables are
      written in single quotes '__'.
      Examples: c__, ctx_or_seq_variable__, vArIaBle1__
  Function and individual variables:
      Identifiers that start with lower case letters and end with a single
      underscore.  Anonymous function and individual variables are
      written in single quotes '_'.
      Examples: i_, fun_or_ind_variable_, vArIaBle1_
  Constants:
       Prolog atoms. Examples: See any textbook on Prolog.
  Reserved words:
       hole, seq. THEY SHOULD NOT BE USED IN THE INPUT PROBLEM.
       star, or, conc. THEY CAN BE USED ONLY AS REGULAR OPERATORS.
  Terms: Built using the grammar:
       T ::= x_ | '_' | x__ | '__' | hole | f(T1,...,Tn) |
             f_(T1,...,Tn) | c__(T)
       where f is a flexible arity function constant, and T in c__(T)
       is not a sequence variable.
       NOTE: Current implementation does not check input for syntactical
       correctness. It is solely user's responcibility to provide a
       syntactically correct input. In the next versions it will be added.
  Contexts: Terms with only one hole.
  Matching Equations:
       E ::= T1 << T2
       where T1 and T2 are not sequence variables and do not contain hole.
  Regular Expressions on Terms:
       Built using the grammar:
       R ::= T | seq([]) | conc(R1,R2) | or(R1,R2) | star(R)
  Regular Expressions on Contexts:
       Built using the grammar:
       Q ::= C | conc(Q1,Q2) | or(Q1,Q2) | star(Q)
       where C is a context.
  Regular Constraints:
       RC ::= [x__, R, 0] | [x__, R, 1] | [c__, Q, 0] | [c__, Q, 1]
       The user can also write [x__, R] for [x__, R, 0] and [c__, Q] for
       [c__, Q, 0].
  Matching problems:
       MP ::= [E1,...,En, RC1,...,RCm].

Output syntax:
  Bindings:
       B ::= x_ ~> T | x__~> seq([T1,...,Tn]) | f_ ~> f | c__ ~> C
       where n>=0, x_ is a individual variable, x__ is a sequence
       variable, f_ is a function variable, c__ is a context variable,
       Ts are terms, f is a function symbol, C is a context. x=!=T,
       seq([x__])=!=seq([T1,...,Tn]), c__(hole)=!=C.
  Substitutions:
       S ::= [B1,...,Bn]
       n>=0.
       
=========================================================================

References to the Regular Context sequence Matching Algorithm:

Temur Kutsia and Mircea Marin. Matching with Regular Constraints.
In: Geoff Sutcliffe and Andrei Voronkov, editors. Proc. of 12th
Int. Conference on Logic in Programming, Artificial Intelligence,
and Reasoning (LPAR'05). December 2-6 2005, Montego Bay, Jamaica.
Volume 3835 of LNAI, pp. 215-229. Springer.

Temur Kutsia and Mircea Marin. Matching with Regular Constraints.
Technical Report 05-05, Research Institute for Symbolic Computation,
Johannes Kepler University of Linz, Austria, 2005.

************************************************************************/

:-dynamic(numb/1).

:-op(120,xfy,~>).
%:-op(110,yf,*).



                /**********************************
                 *    CSMRC: The main predicate   *
                 **********************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
csmrc(MatchingProblem,Matcher) accepts a matching problem MatchingProblem. If
MatchingProblem is solvable, a matching substitution is returned in Matcher.
The user can get all the answers one by one, by pressing the ';' key after
each answer.

csmrc works as follows: First, the terms in the input problem are converted
into lists of symbols. Then a matcher is computed. Next, if the matcher
contains anonymous variables it is recomputed after instantiating anonymous
variables with fresh singleton variables. From the obtained substitution all
bindings for temporary variables (i.e., those variables that do not occur in
MatchingProblem) are removed. Finally, the output is translated into term
substituions.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */



csmrc(MatchingProblem,Matcher):-
   term_matching_problems_to_symbol_list_matching_problems(MatchingProblem,MP),
   asserta(numb(1000)),
   match(MP,[],M),
   recompute_match_with_anonymous_variables_instantiated(MP,M,
                                                         MWithoutAnonymous),
   remove_bindings_for_temporary_variables(MWithoutAnonymous,
                                           SmallerMWithoutAnonymous),
   retractall(numb),
   symbol_list_substitutions_to_term_substitutions(SmallerMWithoutAnonymous,
                                                   Matcher).
                                                   


             /*******************************************************
             *    Conversions from input format to internal form    *
             *******************************************************/

% term_to_list transforms a term into its list representation.
% Example: the term f_(a,g__(x__,h(b)),y_) will be translated into
% [f_, a, [g__, x__, [h, b]], y_]

term_to_list(Var,Term):-var(Var),
        term_to_atom(Var,Term),
        !.
term_to_list(seq([]),_):-
        write('Error. seq([]) is not allowed to occur in matching equations.'),
        nl,
        abort.
term_to_list(Ctx_or_seq_or_fun_or_ind_var_or_constant,X):-
        functor(Ctx_or_seq_or_fun_or_ind_var_or_constant,_,0),
        !,
        transform_var_or_constant(Ctx_or_seq_or_fun_or_ind_var_or_constant,X).
term_to_list(Term,List):-Term=..PseudoList,
        pseudo_list_to_list(PseudoList,List).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Transforms variables into terms of the form ctx_or_seq_var(Var) or
 fun_or_ind_var(Var). Leaves constants unchanged.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


transform_var_or_constant(Fun_Or_Ind_Var,fun_or_ind_var(Fun_Or_Ind_Var)):-
        fun_or_ind_var(Fun_Or_Ind_Var),
        !.
transform_var_or_constant(Ctx_Or_Seq_Var,ctx_or_seq_var(Ctx_Or_Seq_Var)):-
        ctx_or_seq_var(Ctx_Or_Seq_Var),
        !.
transform_var_or_constant(Term,Term):-functor(Term,_,0),
        !.

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Checks for nonanonymous context, sequence, function and individual
 variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ctx_or_seq_var([_|_]):-!,fail.
ctx_or_seq_var(X):-name(X,N), append(_,[_,95,95],N),!.

fun_or_ind_var([_|_]):-!,fail.
fun_or_ind_var(X):-name(X,N), append(_,[C,95],N),!,C\=95.


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 pseudo_list_to_list takes a list of symbols and terms and returns
 a list of symbols and list representations of terms.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

pseudo_list_to_list([],[]).
pseudo_list_to_list([Term|OtherTerms],[TermList|OtherTermsList]):-
       term_to_list(Term,TermList0),
       flatten_comma(TermList0,TermList),
       pseudo_list_to_list(OtherTerms,OtherTermsList).

flatten_comma([',',[','|T1],[','|T2]],Res):-
           !,
           append(T1,T2,R),
           flatten_comma([','|R],Res).

flatten_comma([',',[','|T1],T2],Res):-
           !,
           append(T1,T2,R),
           flatten_comma([','|R],Res).

flatten_comma([',',T1,[','|T2]],[',',T1|R]):-
           !,
           flatten_comma([','|T2],[','|R]).

flatten_comma([',',T1,T2],[',',TR1,TR2]):-
           !,
           flatten_comma(T1,TR1),
           flatten_comma(T2,TR2).

flatten_comma([],[]):-!.
flatten_comma([X|L],[X1|L1]):- !,
           flatten_comma(X,X1),
           flatten_comma(L,L1).

flatten_comma(X,X).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 term_matching_problems_to_symbol_list_matching_problems transforms
 terms in matching problems into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

term_matching_problems_to_symbol_list_matching_problems([],[]).
term_matching_problems_to_symbol_list_matching_problems([S << T|Rest],
                                                        [SL << TL|RestL]):-
        term_to_list(S,SL),
        term_to_list(T,TL),
        term_matching_problems_to_symbol_list_matching_problems(Rest,RestL).

term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp,
                                                          IntegerFlag]|Rest],
                                                        [[ctx_or_seq_var(Var),
                                                          RegexpL,
                                                          IntegerFlag]|RestL]):-
        regexp_to_list(Regexp,RegexpL),
        term_matching_problems_to_symbol_list_matching_problems(Rest,RestL).

term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp]|Rest],L):-
        term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp,
                                                                  0]|Rest],
                                                                L).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 regexp_to_list transforms terms in regular expressions
 into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*  Uncomment this if your Prolog dialect supports postfix * and infix | and ','.
regexp_to_list(R*,star(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list((R1|R2),or(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list((R1,R2),conc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
*/
regexp_to_list(star(R),star(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list(or(R1,R2),or(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(conc(R1,R2),conc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(seq([]),seq([])):-!.
regexp_to_list(Term,TermList):-
        term_to_list(Term,TermList).


        /*******************************************************
        *  Conversions from internal form to the input format  *
        ********************************************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 symbol_list_substitutions_to_term_substitutions transforms
 symbol list representations of terms in substitutions
 into their input format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

symbol_list_substitutions_to_term_substitutions([],[]).
symbol_list_substitutions_to_term_substitutions([X~>TL |Rest1],
                                                [Y~> T|Rest2]):-
        symbol_list_to_term(X,Y),
        symbol_list_to_term(TL,T),
        symbol_list_substitutions_to_term_substitutions(Rest1,Rest2).


symbol_list_to_term(ctx_or_seq_var(Var),Var):-!.
symbol_list_to_term(fun_or_ind_var(Var),Var):-!.
symbol_list_to_term(seq([]),seq([])):-!.
symbol_list_to_term(seq([X1]),X2):-!,
        symbol_list_to_term(X1,X2).
symbol_list_to_term(seq([H1,H2|Tail]),(X,Y)):-!,
        symbol_list_to_term(H1,X),
        symbol_list_to_term(seq([H2|Tail]),Y).
symbol_list_to_term([Head|Tail],Term):-!,
        list_of_symbol_lists_to_symbol_list(Tail,T1),
        Term=..[Head|T1].
symbol_list_to_term(Term,Term).


list_of_symbol_lists_to_symbol_list([],[]).
list_of_symbol_lists_to_symbol_list([X|T],[X|T1]):-
        functor(X,_,0),
        !,
        list_of_symbol_lists_to_symbol_list(T,T1).
list_of_symbol_lists_to_symbol_list([X|T],[X1|T1]):-
        symbol_list_to_term(X,X1),
        list_of_symbol_lists_to_symbol_list(T,T1).


        /*******************************************************
        *   Get substituitions that bind input variables only    *
        ********************************************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_bindings_for_temporary_variables gets a substitution as the input
and removes from it the bindings for all those variables that did not occur
in the input problem but have been generated at run-time.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

remove_bindings_for_temporary_variables([],[]).
remove_bindings_for_temporary_variables([ctx_or_seq_var(Var) ~> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([fun_or_ind_var(Var) ~> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([Var ~> T |Tail],[Var ~> T |Tail1]):-
        remove_bindings_for_temporary_variables(Tail,Tail1).
        


        /*******************************
        *    Transformation rules      *
        *******************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Basic transformation rules
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Success
match([],Acc,Acc).

% Trivial
match([X << X|Tail],Acc,Subst):-!,match(Tail,Acc,Subst).

% Nullary funtions

match([[X|Xs] << X|Tail],Acc,Subst):-
        !,
        match([[X|Xs] << [X]|Tail],Acc,Subst).
match([X << [X]|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular constraints on terms
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% ESRET (Empty sequence in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),seq([]),Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],NewXs),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> seq([])],NewTail),
        match([[X|NewXs] << [X|Ys]|NewTail],NewAcc,Subst).

% CRET (Concatenation in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),conc(R1,R2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,Value),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),
                                              TempSeqVar1),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),
                                              TempSeqVar2),
        new_flags_for_concatenation(Value,TempSeqVar1,NewFlag1,NewFlag2),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([TempSeqVar1,TempSeqVar2])],NewAcc),
        instance_of_term_list(Xs,
                              [ctx_or_seq_var(SeqVar) ~>
                               seq([TempSeqVar1,TempSeqVar2])],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~>
                               seq([TempSeqVar1,TempSeqVar2])],Tail1),
        NewTail=[[TempSeqVar1,R1,NewFlag1],[TempSeqVar2,R2,NewFlag2]|Tail1],
        match([[X,TempSeqVar1,TempSeqVar2|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% ChRET (Choice in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),or(R1,R2),Flag]|L2],Tail),
        !,
        select_branch(or(R1,R2),Rselected),
        evaluate_flag_for_seq_var(Flag,Value),
        append([[ctx_or_seq_var(SeqVar),Rselected,Value]|L1],L2,NewTail),
        match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|NewTail],Acc,Subst).

% RRET1 (Repetition in a regular expression for terms)


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),star(_),Flag]|L2],Tail),
        evaluate_flag_for_seq_var(Flag,0),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> seq([])],NewTail),
        match([[X|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% RRET2 (Repetition in a regular expression for terms)


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),star(R),_]|L2],Tail),
        !,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),
                                              TempSeqVar),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~>
                                  seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],
                              Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~>
                               seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],
                           Tail1),
        NewTail=[[TempSeqVar,R,1],[ctx_or_seq_var(SeqVar),star(R),0]|Tail1],
        match([[X,TempSeqVar,ctx_or_seq_var(SeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

% SVRET (Sequence variable in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),
                    ctx_or_seq_var(AnotherSeqVar),0]|L2],Tail),
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     ctx_or_seq_var(AnotherSeqVar)],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~>
                                  ctx_or_seq_var(AnotherSeqVar)],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~>
                               ctx_or_seq_var(AnotherSeqVar)],NewTail),
        match([[X,ctx_or_seq_var(AnotherSeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),
                    ctx_or_seq_var(AnotherSeqVar),1]|L2],Tail),
        !,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(AnotherSeqVar),
                                              fun_or_ind_var(IndVar)),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([fun_or_ind_var(IndVar),
                          ctx_or_seq_var(AnotherSeqVar)]),
                     ctx_or_seq_var(AnotherSeqVar) ~>
                     seq([fun_or_ind_var(IndVar),
                          ctx_or_seq_var(AnotherSeqVar)])],
                NewAcc),
        instance_of_term_list(Xs,
                              [ctx_or_seq_var(SeqVar) ~>
                               seq([fun_or_ind_var(IndVar),
                                    ctx_or_seq_var(AnotherSeqVar)]),
                               ctx_or_seq_var(AnotherSeqVar) ~>
                               seq([fun_or_ind_var(IndVar),
                                    ctx_or_seq_var(AnotherSeqVar)])],
                              Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,
                           [fun_or_ind_var(SeqVar) ~>
                            seq([fun_or_ind_var(IndVar),
                                 ctx_or_seq_var(AnotherSeqVar)]),
                            ctx_or_seq_var(AnotherSeqVar) ~>
                            seq([fun_or_ind_var(IndVar),
                                 ctx_or_seq_var(AnotherSeqVar)])],
                           NewTail),
        match([[X,fun_or_ind_var(IndVar),
                ctx_or_seq_var(AnotherSeqVar)|Xs1] <<
               [X|Ys]|NewTail],
              NewAcc,Subst).


% TRET (term in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),Term,_]|L2],Tail),
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> Term],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> Term],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> Term],NewTail),
        match([[X,Term|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular expressions on contexts
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% HREC (Hole in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),hole,Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],NewArg),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% CREC (Concatenation in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),conc(Q1,Q2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,Value),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar1),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar2),
        new_flags_for_concatenation(Value,TempCtxVar1,NewFlag1,NewFlag2),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [TempCtxVar1,[TempCtxVar2,hole]]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [TempCtxVar1,[TempCtxVar2,hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar)~>
                               [TempCtxVar1,[TempCtxVar2,hole]]],
                           Tail1),
        NewTail=[[TempCtxVar1,Q1,NewFlag1],[TempCtxVar2,Q2,NewFlag2]|Tail1],
        match([[TempCtxVar1,[TempCtxVar2,NewArg]] << T|NewTail],NewAcc,Subst).

% ChREC (Choice in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),or(Q1,Q2),Flag]|L2],Tail),
        !,
        select_branch(or(Q1,Q2),Rselected),
        evaluate_flag_for_ctx_var(Flag,Value),
        append([[ctx_or_seq_var(CtxVar),Rselected,Value]|L1],L2,NewTail),
        match([[ctx_or_seq_var(CtxVar),Arg] << T|NewTail],Acc,Subst).

% RREC1 (Repetition in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),star(_),Flag]|L2],Tail),
        evaluate_flag_for_ctx_var(Flag,0),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% RREC2 (Repetition in a regular expression for contexts)


match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),star(R),_]|L2],Tail),
        !,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [TempCtxVar,[ctx_or_seq_var(CtxVar),hole]]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [TempCtxVar,[ctx_or_seq_var(CtxVar),hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~>
                               [TempCtxVar,
                                [ctx_or_seq_var(CtxVar),hole]]],
                           Tail1),
        NewTail=[[TempCtxVar,R,1],
                 [ctx_or_seq_var(CtxVar),star(R),0]|Tail1],
        match([[TempCtxVar,
                [ctx_or_seq_var(CtxVar),NewArg]] << T|NewTail],
              NewAcc,Subst).

% CVREC (Context variable in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),
                    [ctx_or_seq_var(AnotherCtxVar),hole],0]|L2],
               Tail),
        !,
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [ctx_or_seq_var(AnotherCtxVar),hole]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [ctx_or_seq_var(AnotherCtxVar),hole]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~>
                               [ctx_or_seq_var(AnotherCtxVar),hole]],
                           NewTail),
        match([[ctx_or_seq_var(AnotherCtxVar),NewArg] << T|NewTail],
              NewAcc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),
                   [ctx_or_seq_var(AnotherCtxVar),hole],1]|L2],
               Tail),
        !,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(AnotherCtxVar),
                                              fun_or_ind_var(FunVar)),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(AnotherCtxVar),
                                              ctx_or_seq_var(SeqVar1)),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(AnotherCtxVar),
                                              ctx_or_seq_var(SeqVar2)),
        compose(Acc,
                [ctx_or_seq_var(CtxVar) ~>
                 [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                  [ctx_or_seq_var(AnotherCtxVar),hole],
                  ctx_or_seq_var(SeqVar2)],
                 ctx_or_seq_var(AnotherCtxVar) ~>
                 [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                  [ctx_or_seq_var(AnotherCtxVar),hole],
                  ctx_or_seq_var(SeqVar2)]],
                NewAcc),
        instance(Arg,
                 [ctx_or_seq_var(CtxVar) ~>
                  [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                   [ctx_or_seq_var(AnotherCtxVar),hole],
                   ctx_or_seq_var(SeqVar2)],
                  ctx_or_seq_var(AnotherCtxVar) ~>
                  [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                   [ctx_or_seq_var(AnotherCtxVar),hole],
                   ctx_or_seq_var(SeqVar2)]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,
                           [ctx_or_seq_var(CtxVar) ~>
                            [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                             [ctx_or_seq_var(AnotherCtxVar),hole],
                             ctx_or_seq_var(SeqVar2)],
                            ctx_or_seq_var(AnotherCtxVar) ~>
                            [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                             [ctx_or_seq_var(AnotherCtxVar),hole],
                             ctx_or_seq_var(SeqVar2)]],
                           NewTail),
        match([[fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                [ctx_or_seq_var(AnotherCtxVar),NewArg],
                ctx_or_seq_var(SeqVar2)] << T|NewTail],
              NewAcc,Subst).

% CRET (Context in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),Ctx,_]|L2],Tail),
        context(Ctx),
        !,
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> Ctx],NewAcc),
        apply_substitution([[ctx_or_seq_var(CtxVar),Arg]<<T],
                           [ctx_or_seq_var(CtxVar) ~> Ctx],NewEquation),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> Ctx],NewTail),
        append(NewEquation,NewTail,NewProblem),
        match(NewProblem,NewAcc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Rules that do not operate on regular expressions
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Sequence variable deletion

match([[X,'__'|Xs] << [X|Ys]|Tail],Acc,Subst):-
        match([[X|Xs] << [X|Ys]|Tail],Acc,Subst).

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],Xs1),
        apply_substitution(Tail,[ctx_or_seq_var(SeqVar) ~> seq([])],Tail1),
        match([[X|Xs1] << [X|Ys]|Tail1],NewAcc,Subst).

% Widening

match([[X,'__'|Xs] << [X,_|Ys]|Tail],Acc,Subst):-
        !,
        match([[X,'__'|Xs] << [X|Ys]|Tail],Acc,Subst).


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([T1,ctx_or_seq_var(SeqVar)])],
                NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~>
                                  seq([T1,ctx_or_seq_var(SeqVar)])],
                              Xs1),
        apply_substitution(Tail,[ctx_or_seq_var(SeqVar) ~>
                                 seq([T1,ctx_or_seq_var(SeqVar)])],
                           Tail1),
        match([[X,ctx_or_seq_var(SeqVar)|Xs1] << [X|Ys]|Tail1],
              NewAcc,Subst).

% Context variable deletion

match([['__',Arg] << T|Tail],Acc,Subst):-
        match([Arg << T|Tail],Acc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        \+(member([ctx_or_seq_var(CtxVar),_RegExp,_Flag],Tail)),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],Arg1),
        apply_substitution(Tail,[ctx_or_seq_var(CtxVar) ~> hole],Tail1),
        match([Arg1 << T|Tail1],NewAcc,Subst).

% Deepening

match([['__',Arg] << [F|Args]|Tail],Acc,Subst):-
        append([F|_],[Proj|_],[F|Args]),
        match([['__',Arg] << Proj|Tail],Acc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << [F|Args]|Tail],Acc,Subst):-
        \+(member([ctx_or_seq_var(CtxVar),_RegExp,_Flag],Tail)),
        append([F|L1],[Proj|L2],[F|Args]),
        append([F|L1],[[ctx_or_seq_var(CtxVar),hole]|L2],Context),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> Context],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> Context],Arg1),
        apply_substitution(Tail,[ctx_or_seq_var(CtxVar) ~> Context],Tail1),
        match([[ctx_or_seq_var(CtxVar),Arg1] << Proj|Tail1],NewAcc,Subst).

% Decomposition

match([[X,S1|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-!,
        match([S1<<T1,[X|Xs] << [X|Ys]|Tail],Acc,Subst).

% Function variable elimination

match([['_'|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        match([[Y|Xs] << [Y|Ys]|Tail],Acc,Subst).

match([[fun_or_ind_var(X)|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[fun_or_ind_var(X) ~> Y],NewAcc),
        instance_of_term_list(Xs,[fun_or_ind_var(X) ~> Y],Xs1),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> Y],Tail1),
        match([[Y|Xs1] << [Y|Ys]|Tail1],NewAcc,Subst).

match([['_'|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        match([[T|Xs] << [T]|Tail],Acc,Subst).

match([[fun_or_ind_var(X)|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        compose(Acc,[fun_or_ind_var(X) ~> T],NewAcc),
        instance_of_term_list(Xs,[fun_or_ind_var(X) ~> T],Xs1),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> T],Tail1),
        match([[T|Xs1] << [T]|Tail1],NewAcc,Subst).

% Individual variable elimination

match(['_' << _Term|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

match([ fun_or_ind_var(X) << Y|Tail],Acc,Subst):-
        !,
        compose(Acc,[fun_or_ind_var(X) ~> Y],NewAcc),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> Y],Tail1),
        match(Tail1,NewAcc,Subst).

match([[_|_]|_],_,_):-
      write('Error. Constrained variable does not occur in matching equations.'),
      nl,abort.


      /*******************************************************************
      *    Recomputing matchers with instantiated anonymous variables   *
      *******************************************************************/

recompute_match_with_anonymous_variables_instantiated(MP,
                                                      OldMatcher,
                                                      NewMatcher):-
        anonymous_to_singleton_variables_in_subst(OldMatcher,
                                                  MatcherWithoutAnonymous),
        recompute_matchers(MP,OldMatcher,MatcherWithoutAnonymous,NewMatcher).


recompute_matchers(_,MatcherWithoutAnonymous,
                   MatcherWithoutAnonymous,MatcherWithoutAnonymous):-!.
recompute_matchers(MP,_,MatcherWithoutAnonymous,NewMatcher):-
        select_equations(MP,Eq),
        apply_substitution(Eq, MatcherWithoutAnonymous,NewMP),
        match(NewMP, MatcherWithoutAnonymous,NewMatcher).


        /*******************************
        *    Check for contexts.       *
        *******************************/


context([hole|Tail]):-!,without_holes(Tail).
context([H|Tail]):-context(H),!,without_holes(Tail).
context([H|Tail]):-without_holes(H),!,context(Tail).

% Check whether a term is without holes, or whether terms in a list
% are all without holes.

without_holes([]):-!.
without_holes([X|Tail]):-without_holes(X),without_holes(Tail).
without_holes(X):-X\=hole.



        /*********************************************
        *        Operations on substitutions         *
        *********************************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
compose(S1,S2,S) composes two substitutions S1 and S2 and returns S.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

compose(S1,S2,S):-
        apply_substitution(S1,S2,S3),
        remove_duplicated_vars_and_append_substs(S3,S2,S).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
apply_substitution takes as input arguments the lists S1 and S2, and returns
S3. S1 may contain bindings, matching equations, regular constraints, and
terms. S2 is a substitution. S3 is obtained from S1 by applying S2 to
terms in it (except variables X in regular constraints [X,R,F]).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

apply_substitution([],_,[]):-!.
apply_substitution(S,[],S):-!.
apply_substitution([X ~> Y |Tail],S,[X ~> Y1| Tail1]):-!,
        instance(Y,S,Y1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([X << Y |Tail],S,[X1 << Y| Tail1]):- !,
        instance(X,S,X1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([[X,Regexp,Flag]|Tail],S,[[X,Regexp1,Flag1]|Tail1]):-!,
        instance(Regexp,S,Regexp1),
        instance(Flag,S,Flag1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([X |Tail],S,[X1 | Tail1]):-
        instance(X,S,X1),
        apply_substitution(Tail, S, Tail1).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
instance(X,Subst,I) returns an instance I of X with respect to the
substitution Subst.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

instance(circplus(X,Y),Subst,circplus(XInst,YInst)):-!,
        instance(X,Subst,XInst),
        instance(Y,Subst,YInst).
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_sequence(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_regexp(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_nonseqvar_term(TermSequence_or_Regexp_or_Term,Subst,
                                   Instance).


% Instance of seq expressions.

instance_of_sequence(seq(L),Subst,seq(Instance)):-
        instance_of_term_list(L,Subst,LInst),
        flatten_sequence(LInst,Instance).


% Instance of regular expressions.

instance_of_regexp(star(R),Subst,star(Instance)):-!,
        instance_of_regexp(R,Subst,Instance).
instance_of_regexp(or(R1,R2),Subst,or(R1Inst,R2Inst)):-!,
        instance_of_regexp(R1,Subst,R1Inst),
        instance_of_regexp(R2,Subst,R2Inst).
instance_of_regexp(conc(R1,R2),Subst,conc(R1Inst,R2Inst)):-!,
        instance_of_regexp(R1,Subst,R1Inst),
        instance_of_regexp(R2,Subst,R2Inst).
instance_of_regexp(seq([]),_,seq([])):-!.
instance_of_regexp(hole,_,hole):-!.
instance_of_regexp([ctx_or_seq_var(CtxVar),hole],Subst,Context):-!,
        instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Context).
instance_of_regexp(ctx_or_seq_var(SeqVar),Subst,Conc):-
 %       SeqVar\=[_|_],        % what is this?
        !,
        instance_of_symbol(ctx_or_seq_var(SeqVar),Subst,Sequence),
        seq_to_conc(Sequence,Conc).
instance_of_regexp(Term,Subst,Instance):-
        instance_of_nonseqvar_term(Term,Subst,Instance).


% Instance of terms that are not sequence variables

instance_of_nonseqvar_term([ctx_or_seq_var(CtxVar),Arg],Subst,Instance):-
        !,
        instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Context),
        instance_of_nonseqvar_term(Arg,Subst,ArgInst),
        apply_context(Context,ArgInst,Instance).

instance_of_nonseqvar_term([FunVar_or_FunSymb|ArgList],Subst,
                           [FunSymb|ArgInst]):-
        !,
        instance_of_symbol(FunVar_or_FunSymb,Subst,FunSymb),
        instance_of_term_list(ArgList,Subst,ArgInst).

instance_of_nonseqvar_term(Symbol,Subst,Instance):-
        instance_of_symbol(Symbol,Subst,Instance).


% Instance of a list of terms.

instance_of_term_list([ctx_or_seq_var(SeqVar)|OtherTerms],Subst,Instance):-
 %       SeqVar \= [_|_],   % do we still need it?
        !,
        instance_of_symbol(ctx_or_seq_var(SeqVar),Subst,Sequence),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst),
        flatten_sequence([Sequence|OtherTermsInst],Instance).

instance_of_term_list([Term|OtherTerms],Subst,
                      [TermInstance|OtherTermsInst]):-
        instance_of_nonseqvar_term(Term,Subst,TermInstance),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst).

instance_of_term_list([],_,[]).


% Instance of a context variable.

instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Instance):-
        member(ctx_or_seq_var(CtxVar)~>Instance,Subst),
        !.
instance_of_ctx_var(ctx_or_seq_var(CtxVar),_,[ctx_or_seq_var(CtxVar),hole]).


% Instance of a symbol.

instance_of_symbol(Var,Subst,Instance):-
        member(Var~>Instance,Subst),
        !.
instance_of_symbol(Constant_or_var_not_in_dom,_,
                   Constant_or_var_not_in_dom).


% Transforming seq expressions into concatenation.

seq_to_conc(seq([X]),X):-!.
seq_to_conc(seq([X,Y|T]),conc(X,R)):-seq_to_conc(seq([Y|T]),R).
seq_to_conc(seq([]),seq([])).


% Context application.

apply_context(hole,Term,Term):-!.
apply_context([hole],Term,Term):-!.
apply_context([Term1,hole],Term,[Term1,Term]):-!.
apply_context([hole|CtxTail],Term,[Term|CtxTail]):-!.
apply_context([[CxtHead|CtxTail]|Tail],Term,[NewTerm|Tail]):-
        apply_context([CxtHead|CtxTail],Term,NewTerm).
apply_context([Head|Tail],Term,[Head|NewTail]):-
        apply_context(Tail,Term,NewTail).

% Flattening nested seq expressions.

flatten_sequence([seq([])|Tail],Tail):-!.
flatten_sequence([seq([X|T])|Tail],[X|Flat]):-!,
        flatten_sequence([seq(T)|Tail],Flat).
flatten_sequence([X|Tail],[X1|Flat]):-!,
        flatten_sequence(X,X1),
        flatten_sequence(Tail,Flat).
flatten_sequence(X,X).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_duplicated_vars_and_append_substs(S1,S2,S) takes two lists of bindings
S1 and S2, deletes all the bindings X~>T from S2 such that S1 contains a
binding for X, deletes all the bindings of the form X~>X from S1, and joins
the obtained lists together into S.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


remove_duplicated_vars_and_append_substs([],L,L):-!.
remove_duplicated_vars_and_append_substs(L,[],L):-!.
remove_duplicated_vars_and_append_substs(Subst1,Subst2,Result):-
        remove_vars_from_second_that_occur_in_first(Subst1,Subst2,
                                                    [],Subst2New),
        remove_identities(Subst1,Subst1New),
        append(Subst1New,Subst2New,Result).
remove_vars_from_second_that_occur_in_first([X~>_|Rest1],[X~>_|Rest2],
                                            ResultSoFar,Result):-
        !,
        remove_vars_from_second_that_occur_in_first(Rest1,Rest2,
                                                    ResultSoFar,Result).
remove_vars_from_second_that_occur_in_first([X~>Y|Rest1],[B|Rest2],
                                            Cleaned,Result):-
        !,
        remove_vars_from_second_that_occur_in_first([X~>Y|Rest1],Rest2,
                                                    [B|Cleaned],Result).
remove_vars_from_second_that_occur_in_first([_|Rest1],[],CleanedOfFirstVar,
                                            Result):-!,
        remove_vars_from_second_that_occur_in_first(Rest1,CleanedOfFirstVar,
                                                    [],Result).
remove_vars_from_second_that_occur_in_first([],Result,_,Result).

remove_identities([],[]).
remove_identities([X~>X|Rest],Result):-
        !,
        remove_identities(Rest,Result).
remove_identities([B|Rest],[B|Result]):-
        remove_identities(Rest,Result).



                 /*******************************
                 *      Various utilities       *
                 *******************************/



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Generating unique temporary context, sequence, function and individual
variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxOrSeqVar),
                                      ctx_or_seq_var(TempCtxOrSeqVar)):-
        create_symbol_list_for_a_temporary_variable(CtxOrSeqVar,UniqueList),
        append(UniqueList,[95,95],List),
        name(TempCtxOrSeqVar,List).

create_unique_temporary_fun_or_indvar(fun_or_ind_var(IndOrFunVar),
                                      fun_or_ind_var(TempIndOrFunVar)):-
        create_symbol_list_for_a_temporary_variable(IndOrFunVar,UniqueList),
        append(UniqueList,[95],List),
        name(TempIndOrFunVar,List).

create_symbol_list_for_a_temporary_variable(SeqVar,UniqueList):-
        name(SeqVar,L),
        remove_last_underscores(L,L1),
        name(temp, TempL),
        append(L1,TempL,TempNameBaseList),
        name(TempNameBase,TempNameBaseList),
        my_gensym(TempNameBase, Unique),
        name(Unique,UniqueList).

remove_last_underscores(L,L1):-append(L1,[95,95],L),!.
remove_last_underscores(L,L1):-append(L1,[95],L).

my_gensym(Base,Unique):-
        numb(N),
        name(Base,L),
        name(N,NL),
        append(L,NL,L1),
        name(Unique,L1),
        N1 is N + 1,
        retract(numb(N)),
        asserta(numb(N1)),!.


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Flag manipulation in regular constraints: Generating new flags for the
constraints that arise after splitting concatenation, and evaluating
flag values.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% New flag generation
new_flags_for_concatenation(0,_,0,0).
new_flags_for_concatenation(1,Var,0,circplus(Var,1)).

% Flag evaluation
evaluate_flag_for_seq_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_seq_var(circplus(X,Y),Value):-
        nonempty_seq(X,V1),
        nonempty_seq(Y,V2),
        my_xor(circplus(V1,V2),Value).

evaluate_flag_for_ctx_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_ctx_var(circplus(X,Y),Value):-
        nonempty_ctx(X,V1),
        nonempty_ctx(Y,V2),
        my_xor(circplus(V1,V2),Value).

/*
my_xor(circplus(X,Y),V):-
        integer(X),
        integer(Y),
        !,
        V is X xor Y.
*/

my_xor(circplus(0,0),1):-!.
my_xor(circplus(0,1),0):-!.
my_xor(circplus(1,0),0):-!.
my_xor(circplus(1,1),0):-!.
my_xor(Unevaluated,Unevaluated).

nonempty_seq(seq([]),0):-!.
nonempty_seq(S,1):-nonempty_seq_aux(S,1),!.
nonempty_seq(S,S).

nonempty_seq_aux(seq([X|_]),1):-X\=ctx_or_seq_var(_),!.
nonempty_seq_aux(seq([_|Y]),V):-nonempty_seq_aux(seq(Y),V).

nonempty_ctx(hole,0).
nonempty_ctx(C,1):-nonempty_ctx_aux(C,1),!.
nonempty_ctx(C,C).

nonempty_ctx_aux([[X|L]|_],1):-nonempty_ctx_aux([X|L],1),!.
nonempty_ctx_aux([[_|_]|Tail],V):-!,nonempty_ctx_aux(Tail,V).
nonempty_ctx_aux([H|_],1):-H\=hole,H\=ctx_or_seq_var(_),!.
nonempty_ctx_aux([_|Tail],V):-nonempty_ctx_aux(Tail,V).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Selecting a branch in the 'or' regular expression
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_branch(or(R1,_R2),R1).
select_branch(or(_R1,R2),R2).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
anonymous_to_singleton_variables_in_subst replaces in a substitution each
anonymous variable with a new temporary singleton variable
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

anonymous_to_singleton_variables_in_subst([],[]).
anonymous_to_singleton_variables_in_subst([X ~> RHS|Tail],[X~>RHS1|Tail1]):-
        anonymous_to_singleton_variables_in_term(RHS,RHS1),
        anonymous_to_singleton_variables_in_subst(Tail,Tail1).

anonymous_to_singleton_variables_in_term('__',Var):-!,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(an__),Var).
anonymous_to_singleton_variables_in_term('_',Var):-!,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(an_),Var).
anonymous_to_singleton_variables_in_term(seq(L),seq(L1)):-!,
        anonymous_to_singleton_variables_in_term(L,L1).
anonymous_to_singleton_variables_in_term([X],[Term]):-!,
        anonymous_to_singleton_variables_in_term(X,Term).
anonymous_to_singleton_variables_in_term([X|Tail],[X1|Tail1]):-!,
        anonymous_to_singleton_variables_in_term(X,X1),
        anonymous_to_singleton_variables_in_term(Tail,Tail1).
anonymous_to_singleton_variables_in_term(X,X).




/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
select_equations takes a matching problem and returns a list of all
matching equations that occur in the input, removing from it all regular
constraints.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_equations([],[]).
select_equations([X << Y|Tail], [X << Y |Tail1]):-!,
        select_equations(Tail,Tail1).
select_equations([_|Tail], L):-select_equations(Tail,L).







